@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2002,2003 Anatoliy Kuznetsov.
+Copyright (c) 2002-2009 Anatoliy Kuznetsov.
 
 Permission is hereby granted, free of charge, to any person 
 obtaining a copy of this software and associated documentation 
@@ -24,15 +24,18 @@ OTHER DEALINGS IN THE SOFTWARE.
 #include <bitset>
 #include <iostream>
 #include <time.h>
+#include <stdio.h>
 
 // No intermix FP with integer SSE in this program
 //#define BM_SET_MMX_GUARD
 //#define BMSSE2OPT
+//#define BMSSE42OPT
+//#define BM64OPT
 
-
-//#define BM_DISBALE_BIT_IN_PTR
 #include "bm.h"
 #include "bmalgo.h"
+#include "bmserial.h"
+//#include "bmdbg.h"
 
 #include <math.h>
 
@@ -43,6 +46,8 @@ const unsigned int REPEATS = 300;
 
 typedef  bitset<BSIZE>  test_bitset;
 
+unsigned platform_test = 1;
+
 
 class TimeTaker
 {
@@ -60,7 +65,14 @@ public:
         clock_t elapsed_clocks = finish_ - start_;
         double duration = (double)(finish_ - start_) / CLOCKS_PER_SEC;
 
-        cout << test_name_ << " ; " <<  elapsed_clocks << ";" << duration << ";";
+        cout << test_name_ << " ; ";
+        if (platform_test) 
+        {
+            cout << duration << endl;
+            return;
+        }
+
+        cout <<  elapsed_clocks << ";" << duration << ";";
         if (repeats_)
         {
             double ops_per_sec = (double)repeats_ / duration;
@@ -173,9 +185,11 @@ void MemCpyTest()
     unsigned* m2 = new unsigned[BSIZE/32];
     
     unsigned int i,j;
+
+    if (!platform_test)
     {
-    TimeTaker tt("Memory ADD transfer test.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("Memory ADD transfer test", REPEATS * 4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         for (j = 0; j < BSIZE/32; j+=4)
         {
@@ -187,9 +201,10 @@ void MemCpyTest()
     }
     }
     
+    if (!platform_test)
     {
-    TimeTaker tt("memcpy transfer test.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("memcpy transfer test", REPEATS * 4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         memcpy(m1, m2, BSIZE/32 * sizeof(unsigned));
     }
@@ -209,21 +224,23 @@ void BitCountTest()
 
     FillSetsIntervals(*bset, *bv, 0, BSIZE, 10);
 
+    if (!platform_test)
     {
-    TimeTaker tt("BitCount. Random bitvector. BM.", REPEATS);
-    for (unsigned i = 0; i < REPEATS; ++i)
+    TimeTaker tt("BitCount. Random bitvector", REPEATS*2);
+    for (unsigned i = 0; i < REPEATS*2; ++i)
     {    
         value+=bv->count();
     }
     }
+
     volatile unsigned* p = &value;
     unsigned c1 = *p;
     c1 = value = 0;
 
-
+    if (!platform_test)
     {
-    TimeTaker tt("BitCount. Random bitvector. STL.", REPEATS);
-    for (unsigned i = 0; i < REPEATS; ++i)
+    TimeTaker tt("BitCount. Random bitvector (STL)", REPEATS*2);
+    for (unsigned i = 0; i < REPEATS*2; ++i)
     {    
         value += bset->count();
     }
@@ -239,6 +256,48 @@ void BitCountTest()
 }
 
 
+void BitForEachTest()
+{
+    // setup the test data
+    //
+    unsigned* test_arr = new unsigned[65536];
+	for (unsigned j = 0; j < 65536; ++j)
+	{
+        test_arr[j] = j;		
+	}
+
+    if (!platform_test)
+    {
+	unsigned bit_list[32];
+    TimeTaker tt("BitList algorithm. Conventional (AND based check)", REPEATS*10);
+    
+
+	for (unsigned i = 0; i < REPEATS*10; ++i)
+    {    
+		for (unsigned j = 0; j < 65536; ++j)
+		{
+			bm::bit_list(i*test_arr[j], bit_list);
+		}
+	}
+	}
+	
+    {
+	unsigned bit_list[32];
+    TimeTaker tt("BitList4 algorithm(sub-octet+switch)", REPEATS*10);
+
+	for (unsigned i = 0; i < REPEATS*10; ++i)
+    {    
+		for (unsigned j = 0; j < 65536; ++j)
+		{
+			bm::bit_list_4(i*test_arr[j], bit_list);
+		}
+	}
+	}
+	
+	delete test_arr;
+}
+
+
 void BitCountSparseTest()
 {
     {
@@ -250,16 +309,17 @@ void BitCountSparseTest()
     SimpleFillSets(*bset, *bv, 0, BSIZE, 2500);
 
     {
-    TimeTaker tt("BitCount. Sparse bitset. BM.", REPEATS);
-    for (unsigned i = 0; i < REPEATS; ++i)
+    TimeTaker tt("BitCount: Sparse bitset ", REPEATS*10);
+    for (unsigned i = 0; i < REPEATS*10; ++i)
     {    
         value += bv->count();
     }
     }
 
+    if (!platform_test)
     {
-    TimeTaker tt("BitCount. Sparse bitset. STL.", REPEATS);
-    for (unsigned int i = 0; i < REPEATS; ++i)
+    TimeTaker tt("BitCount: Sparse bitset (STL)", REPEATS*10);
+    for (unsigned int i = 0; i < REPEATS*10; ++i)
     {    
         value += bset->count();
     }
@@ -272,8 +332,8 @@ void BitCountSparseTest()
     bv->optimize();
 
     {
-    TimeTaker tt("BitCount. GAP Sparse bitset. BM.", REPEATS);
-    for (unsigned i = 0; i < REPEATS; ++i)
+    TimeTaker tt("BitCount: GAP Sparse bitset", REPEATS*1000);
+    for (unsigned i = 0; i < REPEATS*1000; ++i)
     {    
         value += bv->count();
     }
@@ -301,8 +361,8 @@ void BitCompareTest()
     SimpleFillSets(*bset, *bv2, 0, BSIZE, 10);
 
     {
-    TimeTaker tt("BitCompare. Random bitvector. BM.", REPEATS);
-    for (unsigned int i = 0; i < REPEATS; ++i)
+    TimeTaker tt("BitCompare: Random bitvector", REPEATS*10);
+    for (unsigned int i = 0; i < REPEATS*10; ++i)
     {    
         value+=bv1->compare(*bv2);
     }
@@ -314,8 +374,9 @@ void BitCompareTest()
 
     }
 
+    if (platform_test) return;
 
-    unsigned cnt = REPEATS * 50000;
+    unsigned cnt = REPEATS * 100000;
     unsigned* arr1 = new unsigned[cnt];
     unsigned* arr2 = new unsigned[cnt];
 
@@ -334,7 +395,7 @@ void BitCompareTest()
     }
 
     {
-    TimeTaker tt("wordcmp complex. Random words comparison. BM.", cnt);
+    TimeTaker tt("wordcmp complex: Random words comparison", cnt);
 
     for (i = 0; i < cnt; ++i)
     {    
@@ -355,7 +416,7 @@ void BitCompareTest()
     volatile void* p = &c;
 
     {
-    TimeTaker tt("wordcmp0. Random words comparison. BM.", cnt);
+    TimeTaker tt("wordcmp0. Random words comparison", cnt);
     for (i = 0; i < cnt; ++i)
     {    
         c += bm::wordcmp0(arr1[i], arr2[i]);
@@ -364,7 +425,7 @@ void BitCompareTest()
 
 
     {
-    TimeTaker tt("wordcmp. Random words comparison. BM.", cnt);
+    TimeTaker tt("wordcmp. Random words comparison", cnt);
     for (i = 0; i < cnt; ++i)
     {    
         c += bm::wordcmp(arr1[i], arr2[i]);
@@ -376,34 +437,33 @@ void BitCompareTest()
     delete [] arr1;
     delete [] arr2;
 
-
+    char buf[256];
+    sprintf(buf, "%p", p);
 }
 
 void EnumeratorTest()
 {
-    bvect*  bv = new bvect();
+    bvect  bv;
     test_bitset*  bset = new test_bitset();
     unsigned value = 0;
 
-    FillSetsIntervals(*bset, *bv, 0, BSIZE, 10);
+    FillSetsIntervals(*bset, bv, 0, BSIZE, 10);
 
-    unsigned cnt1 = bv->count();
-    unsigned cnt2 = bset->count();
+    unsigned cnt1 = bv.count();
+    //unsigned cnt2 = bset->count();
 
     
     unsigned i;
 
     {
-    TimeTaker tt("Enumerator. BM.", REPEATS);
+    TimeTaker tt("bvector<>::enumerator", REPEATS/10);
     for (i = 0; i < REPEATS/10; ++i)
     {    
-        bvect::enumerator en = bv->first();
-        bvect::enumerator bend = bv->end();
+        bvect::enumerator en = bv.first();
 
-        while (en < bend)
+        for (;en.valid();++en)
         {
             value = *en;
-            ++en;
         }
     }
     }
@@ -411,23 +471,28 @@ void EnumeratorTest()
 
     // -----------------------------------------------
 
+    unsigned cnt = 0;
     {
-    TimeTaker tt("get_next(). BM.", REPEATS/10);
-    for (i = 0; i < REPEATS/10; ++i)
-    {
-        if (bv->count())
+        TimeTaker tt("bvector<>::get_next()", REPEATS/10);
+        for (i = 0; i < REPEATS/10; ++i)
         {
-            unsigned value = bv->get_first();
-            do
+            if (bv.any())
             {
-                value = bv->get_next(value);
-            } while ( value );
+                unsigned value = bv.get_first();
+                do
+                {
+                    value = bv.get_next(value);
+                    cnt += value;
+                } while ( value );
+            }
         }
     }
 
     delete bset;
-    delete bv;
-    }
+//    delete bv;
+
+    char buf[256];
+    sprintf(buf, "%i %i ", cnt, cnt1);//, cnt2);
 
 }
 
@@ -440,13 +505,14 @@ void EnumeratorTestGAP()
     unsigned value = 0;
 
     SimpleFillSets(*bset, *bv, 0, BSIZE, 2500);
+    bv->count();
 
     for (int k = 0; k < 2; ++k)
     {
 
     {
-    TimeTaker tt("Sparse bvector (enumerator). BM.", REPEATS);
-    for (i = 0; i < REPEATS/10; ++i)
+    TimeTaker tt("Sparse bvector (enumerator)", REPEATS*10*(k+1));
+    for (i = 0; i < REPEATS*10*(k+1); ++i)
     {    
         bvect::enumerator en = bv->first();
         bvect::enumerator bend = bv->end();
@@ -462,29 +528,35 @@ void EnumeratorTestGAP()
 
     // -----------------------------------------------
 
+    unsigned cnt = 0;
     {
-    TimeTaker tt("Sparse bvector (get_next). BM.", REPEATS);
+    TimeTaker tt("Sparse bvector (get_next())", REPEATS*10*(k+1));
 
-    for (i = 0; i < REPEATS/10; ++i)
+    for (i = 0; i < REPEATS*10*(k+1); ++i)
     {
-        if (bv->count())
+        if (bv->any())
         {
             unsigned value = bv->get_first();
             do
             {
                 value = bv->get_next(value);
+                cnt += value;
             } while ( value );
         }
     }
-
     }
+    char buf[256];
+    sprintf(buf, "%i", cnt); // to fool some smart compilers like ICC
 
     {
 
     bv->optimize();
     }
 
-    cout << "Testing optimized vectors." << endl;
+    if (!platform_test) 
+    {
+        cout << "Testing optimized vectors." << endl;
+    }
     }
 
     delete bv;
@@ -493,25 +565,88 @@ void EnumeratorTestGAP()
 
 }
 
+void SerializationTest()
+{
+	bvect bv_sparse;
+
+	// prepare a test bitset with a small number of bits set somewhere
+	// far from the beginning
+	for (unsigned i = 0; i < 5; ++i)
+	{
+		bv_sparse[BSIZE/2 + i * 3] = true;		
+	}
+	bv_sparse[100] = true;
+	bv_sparse[70000] = true;
+	bv_sparse[200000] = true;
+
+	bv_sparse.optimize();
+
+	unsigned cnt = bv_sparse.count();
+    bvect::statistics st;
+    bv_sparse.calc_stat(&st);
+    unsigned char*  buf = new unsigned char[st.max_serialize_mem];
+
+    unsigned len, id_size;
+    len = id_size = 0;
+    {
+	TimeTaker tt("Small bvector serialization", REPEATS*70000);
+	for (unsigned i = 0; i < REPEATS*70000; ++i)
+	{
+		len += bm::serialize(bv_sparse, buf, bm::BM_NO_BYTE_ORDER|bm::BM_NO_GAP_LENGTH);
+		id_size += cnt * sizeof(unsigned);
+	}
+	}
+	
+	delete [] buf; buf = 0;
+		
+    bvect*  bv = new bvect();
+    test_bitset*  bset = new test_bitset();
+    unsigned value = 0;
+
+    SimpleFillSets(*bset, *bv, 0, BSIZE, 4);
+    
+	cnt = bv->count();
+    bv->calc_stat(&st);
+    buf = new unsigned char[st.max_serialize_mem];
+    
+    {
+	TimeTaker tt("Large bvector serialization", REPEATS*4);
+	for (unsigned i = 0; i < REPEATS*4; ++i)
+	{
+		len += bm::serialize(*bv, buf, bm::BM_NO_BYTE_ORDER|bm::BM_NO_GAP_LENGTH);
+		id_size += cnt * sizeof(unsigned);		
+	}
+	}
+    
+	char cbuf[256];
+	sprintf(cbuf, "%i %i %i", id_size, len, value);
+    
+    
+    delete bv;
+    delete bset;	
+    delete [] buf;
+}
+
 void InvertTest()
 {
     bvect*  bv = new bvect();
     test_bitset*  bset = new test_bitset();
     unsigned i;
-    unsigned value = 0;
+    //unsigned value = 0;
 
     SimpleFillSets(*bset, *bv, 0, BSIZE, 2500);
     {
-    TimeTaker tt("Invert bvector. BM.", REPEATS/10);
-    for (i = 0; i < REPEATS/10; ++i)
+    TimeTaker tt("Invert bvector", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         bv->flip();    
     }
     }
 
+    if (!platform_test)
     {
-    TimeTaker tt("Invert bvector. STL.", REPEATS);
-    for (i = 0; i < REPEATS/10; ++i)
+    TimeTaker tt("Invert bvector (STL)", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         bset->flip();    
     }
@@ -529,21 +664,22 @@ void AndTest()
     test_bitset*  bset2 = new test_bitset();
     bvect*  bv2 = new bvect();
     unsigned i;
-    unsigned value = 0;
+    //unsigned value = 0;
 
     SimpleFillSets(*bset1, *bv1, 0, BSIZE, 100);
     SimpleFillSets(*bset1, *bv2, 0, BSIZE, 100);
     {
-    TimeTaker tt("AND bvector test. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("AND bvector test", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         *bv1 &= *bv2;
     }
     }
 
+    if (!platform_test)
     {
-    TimeTaker tt("AND bvector test. STL.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("AND bvector test(STL)", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         *bset1 &= *bset2;
     }
@@ -563,19 +699,19 @@ void SubTest()
     test_bitset*  bset1 = new test_bitset();
     bvect*  bv2 = new bvect();
     unsigned i;
-    unsigned value = 0;
 
     SimpleFillSets(*bset1, *bv1, 0, BSIZE, 100);
     SimpleFillSets(*bset1, *bv2, 0, BSIZE, 100);
     delete bset1;
 
     {
-    TimeTaker tt("SUB bvector test. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("SUB bvector test", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         *bv1 -= *bv2;
     }
     }
+        
     delete bv1;
     delete bv2;
 }
@@ -596,10 +732,11 @@ void XorCountTest()
     unsigned count2 = 0;
     unsigned test_count = 0;
 
+    if (!platform_test)
     {
     bvect bv_tmp;
-    TimeTaker tt("XOR COUNT bvector test with TEMP vector. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("XOR COUNT bvector test with TEMP vector", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         bv_tmp.clear(false);
         bv_tmp |= *bv1;
@@ -608,11 +745,11 @@ void XorCountTest()
     }
     }
 
-
+    if (!platform_test)
     {
     test_bitset*  bset_tmp = new test_bitset();
-    TimeTaker tt("XOR COUNT bvector test with TEMP vector. STL.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("XOR COUNT bvector test with TEMP vector (STL)", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         bset_tmp->reset();
         *bset_tmp |= *bset1;
@@ -623,30 +760,35 @@ void XorCountTest()
 
 
     {
-    TimeTaker tt("XOR COUNT bvector test. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("XOR COUNT bvector test", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         count2 += bm::count_xor(*bv1, *bv2);
     }
     }
 
     
-    if (count1 != count2)
-    {
-        cout << "Check failed !" << endl;
-        cout << count1 << " " << count2 << " " << test_count << endl;
-        exit(1);
-    }
+    if (!platform_test)    
+        if (count1 != count2)
+        {
+            cout << "Check failed !" << endl;
+            cout << count1 << " " << count2 << " " << test_count << endl;
+            exit(1);
+        }
     count1 = count2 = 0;
     
     // -----------------------------------------
-    cout << "One optimized vector" << endl;
+    if (!platform_test)
+    {
+        cout << "One optimized vector" << endl;
+    }
     bv2->optimize();
-    //bv2->stat();
+
+    if (!platform_test)
     {
     bvect bv_tmp;
-    TimeTaker tt("XOR COUNT bvector test with TEMP vector. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("XOR COUNT bvector test with TEMP vector", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         bv_tmp.clear(false);
         bv_tmp |= *bv1;
@@ -656,28 +798,33 @@ void XorCountTest()
     }
 
     {
-    TimeTaker tt("XOR COUNT bvector test. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("XOR COUNT bvector test", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         count2 += bm::count_xor(*bv1, *bv2);
     }
     }
 
-    if (count1 != count2)
-    {
-        cout << "Check failed !" << endl;
-        exit(1);
-    }
+    if (!platform_test)
+        if (count1 != count2)
+        {
+            cout << "Check failed !" << endl;
+            exit(1);
+        }
     count1 = count2 = 0;
 
     // -----------------------------------------
-    cout << "Both vectors optimized" << endl;
+    if (!platform_test)
+    {
+        cout << "Both vectors optimized" << endl;
+    }
     bv1->optimize();
     //bv1->stat();
+    if (!platform_test)
     {
     bvect bv_tmp;
-    TimeTaker tt("XOR COUNT bvector test with TEMP vector. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("XOR COUNT bvector test with TEMP vector", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         bv_tmp.clear(false);
         bv_tmp |= *bv1;
@@ -687,17 +834,18 @@ void XorCountTest()
     }
 
     {
-    TimeTaker tt("XOR COUNT bvector test. BM.", REPEATS);
-    for (i = 0; i < REPEATS; ++i)
+    TimeTaker tt("XOR COUNT bvector test", REPEATS*4);
+    for (i = 0; i < REPEATS*4; ++i)
     {
         count2 += bm::count_xor(*bv1, *bv2);
     }
     }
-    if (count1 != count2)
-    {
-        cout << "Check failed !" << endl;
-        exit(1);
-    }
+    if (!platform_test)
+        if (count1 != count2)
+        {
+            cout << "Check failed !" << endl;
+            exit(1);
+        }
     count1 = count2 = 0;
 
 
@@ -724,10 +872,10 @@ void TI_MetricTest()
     unsigned count2 = 0;
     unsigned countA=0, countB=0, test_countA=0, test_countB=0;
     unsigned test_count = 0;
-    double ti1=0, ti2=0, test_ti=0;
+    double ti1=0, ti2=0;
 
     {
-    TimeTaker tt("Tversky Index bvector test vector. BM.", REPEATS);
+    TimeTaker tt("Tversky Index bvector test vector", REPEATS);
     for (i = 0; i < REPEATS; ++i)
     {
         count1 = bm::count_and(*bv1, *bv2);
@@ -740,10 +888,11 @@ void TI_MetricTest()
     }
 
 
+    if (!platform_test)
     {
     test_bitset*  bset_tmp = new test_bitset();
     double test_dice = 0;
-    TimeTaker tt("Dice bvector test with TEMP vector. STL.", REPEATS);
+    TimeTaker tt("Dice bvector test with TEMP vector(STL)", REPEATS);
     for (i = 0; i < REPEATS; ++i)
     {
         bset_tmp->reset();
@@ -768,7 +917,7 @@ void TI_MetricTest()
     dmd[1].metric = bm::COUNT_SUB_AB;
     dmd[2].metric = bm::COUNT_SUB_BA;    
     
-    TimeTaker tt("Tversky Index bvector test (pipeline). BM.", REPEATS);
+    TimeTaker tt("Tversky Index bvector test (pipeline)", REPEATS);
     for (i = 0; i < REPEATS; ++i)
     {
         bm::distance_operation(*bv1, *bv2, &dmd[0], (&dmd[0])+3);
@@ -789,13 +938,16 @@ void TI_MetricTest()
     count1 = count2 = 0;
 
     // -----------------------------------------
-    cout << "One optimized vector" << endl;
+    if (!platform_test)
+    {
+        cout << "One optimized vector" << endl;
+    }
     bv2->optimize();
     bv1->count(); // trying to fool the CPU cache
 
     
     {
-    TimeTaker tt("Dice metric bvector test. BM.", REPEATS);
+    TimeTaker tt("Dice metric bvector test", REPEATS);
     for (i = 0; i < REPEATS; ++i)
     {
         count1 = bm::count_and(*bv1, *bv2);
@@ -815,7 +967,7 @@ void TI_MetricTest()
     dmd[1].metric = bm::COUNT_SUB_AB;
     dmd[2].metric = bm::COUNT_SUB_BA;    
     
-    TimeTaker tt("Tversky Index bvector test (pipeline). BM.", REPEATS);
+    TimeTaker tt("Tversky Index bvector test(pipeline)", REPEATS);
     for (i = 0; i < REPEATS; ++i)
     {
         bm::distance_operation(*bv1, *bv2, &dmd[0], (&dmd[0])+3);
@@ -837,11 +989,14 @@ void TI_MetricTest()
     count1 = count2 = 0;
 
     // -----------------------------------------
-    cout << "Both vectors optimized" << endl;
+    if (!platform_test)
+    {
+        cout << "Both vectors optimized" << endl;
+    }
     bv1->optimize();
 
     {
-    TimeTaker tt("Tversky index bvector test. BM.", REPEATS);
+    TimeTaker tt("Tversky index bvector test", REPEATS);
     for (i = 0; i < REPEATS; ++i)
     {
         count1 = bm::count_and(*bv1, *bv2);
@@ -859,7 +1014,7 @@ void TI_MetricTest()
     dmd[1].metric = bm::COUNT_SUB_AB;
     dmd[2].metric = bm::COUNT_SUB_BA;    
     
-    TimeTaker tt("Tversky Index bvector test (pipeline). BM.", REPEATS);
+    TimeTaker tt("Tversky Index bvector test (pipeline)", REPEATS);
     for (i = 0; i < REPEATS; ++i)
     {
         bm::distance_operation(*bv1, *bv2, &dmd[0], (&dmd[0])+3);
@@ -885,17 +1040,180 @@ void TI_MetricTest()
     delete bset2;    
 }
 
+#ifdef BMSSE2OPT
+# if(_MSC_VER)
+#  define BM_ALIGN16 __declspec(align(16))
+#  define BM_ALIGN16ATTR
+
+# else // GCC
+
+#  define BM_ALIGN16
+#  define BM_ALIGN16ATTR __attribute__((aligned(16)))
+# endif
+
+#else
+
+#  define BM_ALIGN16
+#  define BM_ALIGN16ATTR
+
+#endif 
+
+void BitBlockTransposeTest()
+{
+    bm::word_t BM_ALIGN16 block1[bm::set_block_size] BM_ALIGN16ATTR = {0,};
+    //bm::word_t BM_ALIGN16 block2[bm::set_block_size] = {0xFF,};
+    unsigned   BM_ALIGN16 tmatrix1[32][bm::set_block_plain_size] BM_ALIGN16ATTR;
+
+    for (unsigned i = 0; i < bm::set_block_size; ++i)
+    {
+        block1[i] = 1 | (1 << 5) | (7 << 15) | (3 << 22);
+    }
+
+    const unsigned blocks_count = 70000;
+    bm::word_t* blocks[blocks_count];
+    for (unsigned k = 0; k < blocks_count; ++k)
+    {
+        blocks[k] = bm::block_allocator::allocate(bm::set_block_size, 0);
+        for (unsigned i = 0; i < bm::set_block_size; ++i)
+        {
+            blocks[k][i] = 1 | (1 << 5) | (7 << 15) | (3 << 22);
+        }
+    }
+
+    unsigned cnt=0;
+/*
+    {
+    TimeTaker tt("Bit-block transpose.", REPEATS*1000);
+    for (unsigned i = 0; i < REPEATS*1000; ++i)
+    {
+        bm::bit_block_transpose(block1, tmatrix1);
+    }
+    }
+
+    {
+    TimeTaker tt("Bit-block trestore.", REPEATS*1000);
+    for (unsigned i = 0; i < REPEATS*1000; ++i)
+    {
+        bm::bit_block_trestore(tmatrix1, block2);
+        cnt += block2[10];
+
+    }
+    }
+
+    {
+    TimeTaker tt("Bit-block transpose distance.", REPEATS*1000);
+    unsigned distance[bm::set_block_plain_cnt][bm::set_block_plain_cnt];
+    for (unsigned i = 0; i < REPEATS*1000; ++i)
+    {
+        bm::bit_block_tmatrix_distance(tmatrix1, distance);
+        cnt += distance[1][1];
+    }
+    }
+    printf("", cnt);
+*/
+
+    unsigned d2[bm::set_block_plain_cnt][bm::set_block_plain_cnt];
+    {
+    TimeTaker tt("Bit-block transpose+distance", 100000);
+    unsigned distance[bm::set_block_plain_cnt][bm::set_block_plain_cnt];
+    unsigned idx = 0;
+    for (unsigned i = 0; i < 100000; ++i)
+    {
+        bm::vect_bit_transpose<unsigned, 
+                               bm::set_block_plain_cnt, 
+                               bm::set_block_plain_size>
+                               (blocks[idx], bm::set_block_size, tmatrix1);
+        bm::tmatrix_distance<unsigned, 
+                             bm::set_block_plain_cnt, 
+                             bm::set_block_plain_size>
+                             (tmatrix1, distance);
+    
+        cnt += distance[1][1];
+        ++idx;
+        if (idx >= blocks_count) idx = 0;
+        memcpy(d2, distance, sizeof(distance));
+    }
+
+    }
+    
+    char cbuf[256];
+    sprintf(cbuf, "%i %i", cnt, d2[10][10]);
+
+    for (unsigned i = 0; i < blocks_count; ++i)
+    {
+        bm::block_allocator::deallocate(blocks[i], 0);
+    }
+
+}
+
+void ptest()
+{
+    bvect*  bv_small = new bvect(bm::BM_GAP);
+    bvect*  bv_large = new bvect(bm::BM_GAP);
+
+    test_bitset*  bset = new test_bitset();
+
+    FillSetsIntervals(*bset, *bv_large, 0, 2000000000, 10);
+
+    for (int i = 0; i < 2000; ++i)
+    {
+        bv_small->set(i*10000);
+    }
+
+    {
+    TimeTaker tt("Operation &= test", REPEATS * 10);
+    unsigned count = 0;
+    for (unsigned i = 0; i < REPEATS*10; ++i)
+    {
+        bvect t1(bm::BM_GAP);
+        t1 = *bv_small;
+        t1 &= *bv_large;
+        count += t1.count();
+    }
+    }
+
+
+
+    {
+    TimeTaker tt("Operation &= with enumerator test", REPEATS * 10);
+    unsigned count = 0;
+    for (unsigned i = 0; i < REPEATS*10; ++i)
+    {
+        bvect t1(bm::BM_GAP);
+        bvect t2(bm::BM_GAP);
+        t1 = *bv_small;
+        
+        for (bvect::enumerator it = t1.first(); it != t1.end(); ++it) {
+            if ((*bv_large)[*it]) {
+                t2.set_bit(*it);
+            }
+        }
+        count += t2.count();
+    }
+    }
+
+
+}
+
 
 int main(void)
 {
+//    ptest();
+
+    TimeTaker tt("TOTAL", 1);
 
     MemCpyTest();
 
     BitCountTest();
 
+    BitForEachTest();
+
     BitCountSparseTest();
 
     BitCompareTest();
+
+    BitBlockTransposeTest();
+
     EnumeratorTest();
     EnumeratorTestGAP();
 
@@ -907,6 +1225,13 @@ int main(void)
     XorCountTest();
 
     TI_MetricTest();
+
+    SerializationTest();
         
     return 0;
 }
+
+
+
+
+
